/*
Copyright 2004 Cesar Tardaguila & Javier Tardaguila

This file is part of Drops.

Drops is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Srops is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Drops; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import net.designnation.data.*

class net.designnation.data.List implements IList
{                      
	private var head: Node;
	private var tail: Node;	
	private var elementCount: Number;
	private var comparator: IOrder= undefined;
	
	function List( orderFunctor: IOrder )//Tipear par�metro
	{
		this.head 			= undefined;
		this.tail			= undefined;
		this.elementCount	= 0;
		
		if( orderFunctor!= undefined )
		{
			this.comparator= orderFunctor;		
		}
	}
	
	public function isEmpty(  ): Boolean
	{
		return this.head == undefined;
	}
	
	public function makeEmpty(  ): Void
	{
		this.elementCount = 0;
		this.head = undefined;
	}
	
	//Interface methods
	public function push( element: Object )
	{
		var newNode: Node = new Node( element );
		var currentNode: Node;
		
		if( isEmpty(  ) )
		{
			this.head = newNode;
			this.tail = newNode;
			this.elementCount++;
			//trace( "this.head " + this.head );
		}
		else
		{
			this.tail.next= newNode;
			this.tail	 = newNode;
						
			this.elementCount++;
		}	
	}
	
	public function insertOrdered( element: Object )
	{
		var newNode: Node = new Node( element );	
		var currentNode, prevNode: Node;	
		
		if( isEmpty(  ) )
		{
			this.head = newNode;		
			this.tail = newNode;
		}
		else
		{
			var breaked = false;
			
			currentNode = this.head;
			prevNode	= this.head;
			
			while( currentNode.next!= undefined )
			{
				prevNode	= currentNode;
				currentNode = currentNode.next;				
				
				if( ( this.comparator!= undefined && 
					this.comparator.lessThan( currentNode.next.value, newNode.value ) ) )
				{
					prevNode.next	= newNode;
					newNode.next 	= currentNode;
										
					breaked = true;
					break;
				}
			}
			if( !breaked )
			{
				currentNode.next = newNode;
				this.tail	  		 = newNode;
			}
		}
		this.elementCount++;
	}
		
	
	public function insertElementAt( index: Number, element: Object ): Boolean
	{
		var retVal: Boolean;
		var currentNode: Node;
		
		if( index>= this.elementCount || index< 0 )
		{
			retVal= false;
		}
		else
		{
			var newNode:Node = new Node( element );
			
			this.elementCount++;
			
			currentNode = this.head;			
			for( var k = 0; k< index; k++ )
			{
				currentNode = currentNode.next;
			}
			
			if( this.tail=== currentNode )
			{
				this.tail= newNode;
			}
			newNode.next 	= currentNode.next;
			currentNode.next= newNode;
			
			retVal= true;
		}
		
		return retVal;
	}
	
	public function pop(  ): Object
	{
		if( !isEmpty(  ) )
		{
			this.elementCount--;
			
			if( this.tail== this.head )
			{
				this.tail= undefined;
			}
			
			var returnElement = this.head.value;
			
			this.head = this.head.next;
			
			return returnElement;
		}
	
	}
	
	public function getElementAt( index: Number ): Object
	{
		var currentNode: Node;
		var retVal: Object;
		
		if( index>= this.elementCount || index< 0 )
		{
			retVal= undefined;
		}
		else
		{
			currentNode= this.head;
			
			for( var k= 0; k< index; k++ )
			{				
				currentNode= currentNode.next;
			}			
			retVal= currentNode.value;
		}
		return retVal;
	}
	
	public function deleteElement( element: Object ): Boolean
	{
		var retVal: Boolean= false;
		var currentNode, prevNode: Node;
		
		prevNode= this.head;
				
		for( currentNode= this.head; currentNode!= undefined; currentNode= currentNode.next )
		{						
			if( equals( element, currentNode ) )
			{				
				if( this.tail=== currentNode )
				{
					this.tail= prevNode;
					this.tail.next= null;
				}
				if( this.head=== currentNode )
				{
					this.head= currentNode.next;
					currentNode.next= null;
				}
				
				if( this.head!== currentNode && this.tail!== currentNode )
				{
					prevNode.next= currentNode.next;					
				}
											
				this.elementCount--;
				retVal= true;
				break;
			}
			prevNode= currentNode;
		}
		return retVal;
	}
	
	private function equals( element: Object, currentNode: Node ): Boolean
	{
		var bEq: Boolean= false;
			
		if( this.comparator!= undefined ) 
		{				
			 bEq= this.comparator.equals( currentNode.value, element );
		}
		else if( element.id!= undefined && currentNode.value.id!= undefined )
		{
			bEq= element.id== currentNode.value.id;
		}
		else if( element.id== undefined && currentNode.value.id== undefined )
		{
			bEq= element.toString(  )== currentNode.value.toString(  );	
		}
		
		return bEq;
	}
	
	public function deleteElementAt( index: Number ): Boolean
	{
		var retVal: Boolean= false;
		var currentNode, prevNode: Node;
		
		if( index>= this.elementCount || index< 0 )
		{
			retVal= false;
		}
		else
		{
			this.elementCount--;
			currentNode = this.head;
			prevNode	= this.head;
			
			for( var i= 0; i< index; i++ )
			{
				prevNode	= currentNode;
				currentNode = currentNode.next;				
			}
			
			if( this.tail=== currentNode )
			{
				this.tail= prevNode;
			}
			if( this.head=== prevNode )
			{
				this.head= prevNode.next;
			}
			
			prevNode.next= currentNode.next;
			
			retVal= true;
		}	
		return retVal;
	}
	
	public function exists( element: Object ): Number
	{
		var indice: Number  	= -1;
		var currentNode: Node = this.head;
		
		while( currentNode.next!= undefined )
		{
			indice++;
			currentNode = currentNode.next;
			//trace( fgObjectUtils.toCompare( currentNode.value ) );			
			if( equals( element, currentNode ) )			
			{
				break;
			}
		}
		return indice;
	}
	
	public function toString(  ): String
	{
		var returnStr: String= "List("+ this.elementCount+ " elements):";
		
		var currentNode: Node = this.head;
		
		while( currentNode!= undefined )
		{
			var sAux: String= String( currentNode.value.id );
			
			if( sAux== String( undefined ) )
			{
				sAux= currentNode.value.toString(  );
			}
			
			returnStr+= sAux +"\n";
			
			currentNode = currentNode.next; 	
		}
		return returnStr;
	}
	
	public function count(  ): Number
	{
		return elementCount;
	}
	
	public function iterator(  ): IIterator
	{
		return new ListIteratorImpl( this );
	}	
}